scriptname sslRapeSpell extends ReferenceAlias
; The version number for SexLabRapeSpell
int version = 6

; Load the SexLab framework for use
SexLabFramework property SexLab auto

; Actor selection storage
Actor[] Slots


; Trigger lock to make sure we don't accidentally call TriggerSex() more than once
; on the same set of actors after another's spell effect expires.
bool locked

; Our spells
Spell property SexLabRapeSpellTarget auto
Spell property SexLabRapeSpellBuff auto

Faction Property SSLRSLoverFaction auto
Spell Property SSLRSCriminalPulseSpell auto
Spell Property SSLRSVictimDrainSpell auto
Spell Property SSLRSAggressDrainSpell auto
Spell Property SSLRSFearSpell auto
Spell Property SSLRSCalmSpell auto

ReferenceAlias Property PlayerRef auto

Perk Property SSLRSSexualInferiorityPerk auto

Actor Property caster = None auto
Actor Property target = None auto

int previousCrimeGold = 0
float previousArousal = 0.0
bool interrupted = false
bool wasProtected = false
bool wasNaked = false

; Initializes the RapeSpell script back to a state with no registered actors
function ResetSlots()
	; Empty out the Slots array by initializing it to an empty array.
	Actor[] emptyslots
	Slots = emptyslots
	; Unlock our TriggerSex() function for future use.
	locked = false
endFunction

; Called on first startup with mod enabled
event OnInit()
	; Give player the irrestible aura spells that trigger everything here.
	Game.GetPlayer().AddSpell(SexLabRapeSpellTarget, true)
	; Game.GetPlayer().AddSpell(RapeSpellSelf, true)
	; Init slots and locked status
	ResetSlots()
	if (self != PlayerRef)
		clear()
	endIf
endEvent

; Reset registered actors and lock whenever player reloads their game.
; Ensures we start out in a useable state on every load; just in case.
event OnPlayerLoadGame()
	if version < 6
		SexLab.PlayerRef.RemoveSpell(SexLabRapeSpellBuff)
		version = 6
	endIf
	ResetSlots()
endEvent

; Utility function to check if the actor is slotted already or not
; Returns true if they are in the Slots array already
bool function IsSlotted(actor Position)
	return Slots.Find(Position) != -1
endFunction

; Called from spell, validates the target and places actor in storage for scene use.
; Returns True if the actor is successfully added to storage, False if they fail to validate.
bool function RegisterActor(actor Position, Actor Aggressor, bool isTarget)
	interrupted = false
	; Validate we can add the actor before trying to register them.

	; IsSlotted() will check if they are already registered or not.
	if IsSlotted(Position)
		; Position is already slotted, fail and show some flavor text.
		showMessage(Position.GetLeveledActorBase().GetName() + " is already in action!")
		return false ; They didn't pass validation; don't continue with this function.

	; Check and make sure we aren't already over our limit of 3 actors
	; Also checks if they are a valid actor to animate in the eyes of SexLab
	; Slots.Length >= 3 makes sure our Slots array has room for them with a max of 3.
	; ValidateActor() will return a negative number if they are rejected.
	elseIf Slots.Length >= 2 || SexLab.ValidateActor(Position) < 1
		; Flavor text for the failure.
		showMessage("Your spell fizzles and fails to affect " + Position.GetLeveledActorBase().GetName())
		return false ; They didn't pass validation; don't continue with this function.

	endIf

	; Still here? The actor has passed our validation checks then.
	if (Position == Aggressor)
		caster = Position
	endIf
	if (isTarget)
		target = Position
	endIf

	; Slot the actor into our registration array.
	Slots = sslUtility.PushActor(Position, Slots)
	; Flavor text for the successful cast
	showMessage(Position.GetLeveledActorBase().GetName() + " gets involved in action!")

	; If we have our limit of 3 actors now, lets not wait and just start the scene now.
	if Slots.Length >= 2
		TriggerSex()
	endIf
	return true
endFunction

; Takes our current Slots array and attempts to start a sex scene using the contained actors, however many actors it has, between 1 and 3.
function TriggerSex()

	; Ensure we aren't triggering a second time from another actors spell effect expiring while we are still processing this one
	; Also make sure we even have the actors to do anything first.
	if locked || Slots.Length == 0
		return ; TriggerSex() is already processing OR no actor slots filled. Stop now and don't trigger any sex.
	endIf

	; Set our lock to prevent any calls to TriggerSex() on our current actor slots.
	locked = true

	; Claim an available animation thread from SexLab we can add our actors to.
	sslThreadModel Thread = SexLab.NewThread()

	; All animations so far in SexLab expect the female to the first actor given to it.
	; So make sure we have the female actors in front of the array so animation positions are properly set
	; SortActors() from SexLab will return the given array with females moved to the front of the array by default.
	Slots = SexLab.SortActors(Slots)

	bool aggressive = true
	if (caster.getLeveledActorBase().getSex() == 1 && target.getLeveledActorBase().getSex() == 0)
		; Random choice for assault from female to male
		aggressive = (Utility.randomInt(0, 1) == 1)
		if (!aggressive)
			stopAlarm()
		endIf
	endIf

	; Loop through registered actors to the thread from our now sorted Slots array.
	int i = 0
	while i < Slots.Length
		; Add looped actor to the thread. AddActor() will return -1 if they failed to be added to the thread.
		if Thread.AddActor(Slots[i], (Slots[i] == target) && aggressive) == -1
			; The actor failed to be added to the thread for some reason, so lets not continue.
			; Send error message to the debug log.
			Debug.Trace("--- SexLab RapeSpell --- Failed to slot the actor '" + Slots[i].GetLeveledActorBase().GetName() + "' into the animation thread["+Thread.tid+"]!")
			; Unlock the animation thread so it clears itself and doesn't try to continue with any actors we may have successfully added.
			; Thread.UnlockThread()
			; Clear our actors & unlock our TriggerSex() function for future use.
			ResetSlots()
			return ; Stop the function now since we failed to add the actor.
		endIf
		if (Slots[i] == target)
			Thread.disableRagdollEnd(Slots[i], false)
		else
			Thread.disableRagdollEnd(Slots[i], true)
		endIf
		i += 1
	endWhile

	; Set our custom hook name "RapeSpell" in the thread.
	Thread.SetHook("RapeSpell"+getName())
	; Register our custom event "EndBuff" onto the AnimationEnd hook, which is triggered once
	; the actors have finished and begun to reset.
	; _RapeSpell makes it a non global hook that is called from our above Thread.SetHook("RapeSpell"),
	; so only this scene will call the EndBuff event and not other animations started outside of RapeSpell.
	RegisterForModEvent("AnimationStart_RapeSpell"+getName(), "AnimationChanged")
	RegisterForModEvent("LeadInEnd_RapeSpell"+getName(), "AnimationChanged")
	RegisterForModEvent("StageStart_RapeSpell"+getName(), "periodicReport")
	RegisterForModEvent("OrgasmStart_RapeSpell"+getName(), "periodicReport")
	RegisterForModEvent("AnimationChange_RapeSpell"+getName(), "AnimationChanged")
	RegisterForModEvent("AnimationEnd_RapeSpell"+getName(), "EndBuff")

	; We'll keep the scene basic and let SexLab and player configuration figure out most of the stuff on it's own.
	; So let's start the animation now.
	thread_subject =Thread.StartThread()

	; We are done with our registered actors, so reset our lock and clear out our registered actors.
	ResetSlots()

endFunction

sslThreadController thread_subject = None
string flatten = ""
event AnimationChanged(string eventName, string argString, float argNum, form sender)
	target.dispelSpell(SSLRSVictimDrainSpell)
	caster.dispelSpell(SSLRSAggressDrainSpell)
	sslBaseAnimation sba = thread_subject.Animation
	string [] tags = sba.getTags()
	string [] skills = new string [3]
	skills[0] = "vaginal"
	skills[1] = "anal"
	skills[2] = "oral"
	flatten = ""
	int i = 0
	float caster_level = 1.0
	float target_level = 1.0
	while (i < tags.Length)
		int idx = skills.find(tags[i])
		if (idx >= 0)
			caster_level += SexLab.Stats.getSkillLevel(caster, skills[idx]) as float
			target_level += SexLab.Stats.getSkillLevel(target, skills[idx]) as float
			if (flatten != "")
				flatten += "/"
			endIf
			flatten += skills[idx]
		endIf
		i += 1
	endWhile
	if (flatten == "")
		flatten = "Uncategorized"
	endIf
	i = 0
	float C = 0.011
	float pF = 0.75
	float rF = 1 - pF
	float victimBaseMag = C*math.pow(strongerMaxValue(target), pF)*math.pow(strongerMaxValue(caster), rF)
	float aggressBaseMag = C*math.pow(strongerMaxValue(caster), pF)*math.pow(strongerMaxValue(target), rF)
	float victimDamage = victimBaseMag * (pF*caster_level+rF*target_level)/(caster_level+target_level)
	float aggressDamage = aggressBaseMag * (pF*target_level+rF*caster_level)/(caster_level+target_level)
	if (victimDamage && aggressDamage)
		while (i < SSLRSVictimDrainSpell.getNumEffects())
			SSLRSVictimDrainSpell.setNthEffectMagnitude(i, victimDamage)
			SSLRSAggressDrainSpell.setNthEffectMagnitude(i, aggressDamage)
			i += 1
		endWhile
	endIf
	SSLRSVictimDrainSpell.cast(target, target)
	SSLRSAggressDrainSpell.cast(target, caster)
	showMessage(flatten+" sex - "+caster.getLeveledActorBase().getName()+": Level "+caster_level as int+" / "+target.getLeveledActorBase().getName()+": Level "+target_level as int)
	periodicReport(eventName, argString, argNum, sender)
endEvent

float Function strongerMaxValue(Actor akActor)
	float tsta = akActor.getActorValue("Stamina")
	float tstap = akActor.getActorValuePercentage("Stamina")
	if (tstap)
		tsta /= tstap
	endIf
	float tmag = akActor.getActorValue("Magicka")
	float tmagp = akActor.getActorValuePercentage("Magicka")
	if (tmagp)
		tmag /= tmagp
	endIf
	float tstrong = tsta
	if (tstrong < tmag)
		tstrong = tmag
	endIf
	if (tstrong == 0.0)
		tstrong = (akActor.getBaseActorValue("Stamina")+akActor.getBaseActorValue("Magicka"))/2
	endIf
	return tstrong
endFunction

float Function strongerValue(Actor akActor)
	float tsta = akActor.getActorValue("Stamina")
	float tmag = akActor.getActorValue("Magicka")
	float tstrong = tsta
	if (tstrong < tmag)
		tstrong = tmag
	endIf
	return tstrong
endFunction

function periodicReport(string eventName, string argString, float argNum, form sender)
	bool friendly = target.getRelationshipRank(caster) > 0
	if (friendly || !caster.isDetectedBy(target) || wasNaked || target == Game.getPlayer() || caster.isGuard())
		; nothing
	else
		SSLRSCriminalPulseSpell.cast(caster, target)
	endIf
	float victimDamage = SSLRSVictimDrainSpell.getNthEffectMagnitude(0)
	float aggressDamage = SSLRSAggressDrainSpell.getNthEffectMagnitude(0)
	float tstrong = strongerValue(target)
	float ttime = tstrong/victimDamage
	float cstrong = strongerValue(caster)
	float ctime = cstrong/aggressDamage
	if (math.abs(ctime/ttime - 1.0) <= 0.1)
		showMessage("This is going to be a fair bout")
	elseIf (ctime > ttime)
		if (ctime > ttime*2)
			showMessage(caster.getLeveledActorBase().getName()+" is overwhelming in this position")
		else
			showMessage(caster.getLeveledActorBase().getName()+" has advantage in this position")
		endIf
	else
		if (ttime > ctime*2)
			showMessage(target.getLeveledActorBase().getName()+" is overwhelming in this position!")
		else
			showMessage(target.getLeveledActorBase().getName()+" has advantage in this position!")
		endIf
	endIf
	if (ttime as int == 0)
		showMessage(target.getLeveledActorBase().getName()+" is already exhausted")
	else
		showMessage(target.getLeveledActorBase().getName()+" will be exhausted in "+(ttime as int)+" seconds")
	endIf
	if (caster != Game.getPlayer())
		bool leadin = thread_subject.Animation.HasTag("LeadIn")
		float x = ((cstrong/strongerMaxValue(caster))+(tstrong/strongerMaxValue(target)))/2.0
		if (x > 1.0)
			x = 1.0
		endIf
		float stageFactor = 0.5 + 0.5*thread_subject.Stage/thread_subject.Animation.StageCount
		float y = stageFactor*1.2*x*(1.5 - math.pow(x, 3))
		y -= 0.1  ; cutoff
		;debug.notification("backward threshold["+thread_subject.Stage+"]: "+y)
		if (!leadin && thread_subject.Stage > 2 && Utility.randomFloat(0, 1.0) <= y)
			thread_subject.advanceStage(true)
		endIf
	endIf
endFunction

string function format(float value)
	int integerPart = (value as int)*10
	int decimalPart = (value*10) as int
	decimalPart = decimalPart - integerPart
	integerPart /= 10
	return (integerPart as string)+"."+(decimalPart as string)
endFunction

; Our buff hook, called from RegisterForModEvent("AnimationEnd_RapeSpell", "EndBuff")
; AnimationEnd is sent by SexLab called once the sex animation has fully stopped
event EndBuff(string eventName, string argString, float argNum, form sender)

	; Get our list of actors at the end
	Actor[] Positions = SexLab.HookActors(argString)

	; Loop through our list of actors
	int i = Positions.Length
	while i
		i -= 1
		; Flavor text for end scene in place of the removed buff
		; Debug.Notification(Positions[i].GetLeveledActorBase().GetName() + "'s irrestible aura fades.")
		; Removed in Rev 6 due to bug with the buff spell
		; Add our buff spell to the looped actor
		; Positions[i].AddSpell(SexLabRapeSpellBuff, false)
		Positions[i].RemoveSpell(SexLabRapeSpellBuff)
	endWhile

	target.dispelSpell(SSLRSVictimDrainSpell)
	caster.dispelSpell(SSLRSAggressDrainSpell)
	bool casterWeak = ((caster.GetActorValue("Magicka") == 0 || caster.GetActorValuePercentage("Magicka") < 0.1) && (caster.GetActorValue("Stamina") == 0 || caster.GetActorValuePercentage("Stamina") < 0.1))
	bool targetWeak = ((target.GetActorValue("Magicka") == 0 || target.GetActorValuePercentage("Magicka") < 0.1) && (target.GetActorValue("Stamina") == 0 || target.GetActorValuePercentage("Stamina") < 0.1))
	bool friendly = target.getRelationshipRank(caster) > 0
	bool satisfied = previousArousal - getArousal(target) >= 75

	if (friendly || (!interrupted && targetWeak) || satisfied)
		; note that 'satisfied' is option exclusive to SexLab Aroused addon. Avoid sole clause of 'satisfied'.
		; accept
		;; crime alarm withdrawal
		if (targetWeak || satisfied)
			debug.sendAnimationEvent(target, "BleedoutStart")
			registerForSingleUpdate(8)
			stopAlarm()
			if (caster == Game.getPlayer())
				;;; avoid player cast (causes unintentional level up)
				SSLRSCalmSpell.cast(target, target)
			else
				SSLRSCalmSpell.cast(caster, target)
			endIf
			;; relationship advancement
			if (!friendly && !casterWeak)
				target.setRelationshipRank(caster, target.getRelationshipRank(caster)+1)
				showMessage("Relation with "+target.getLeveledActorBase().getName()+" improved")
				friendly = true
			endIf
		endIf
		;; yielding
		if (friendly && targetWeak && !casterWeak)
			occupyCell(caster, target)
		endIf
	else
		; decline
		if (!friendly && targetWeak)
			if (caster == Game.getPlayer())
				;;; avoid player cast (causes unintentional level up)
				SSLRSFearSpell.cast(target, target)
			else
				SSLRSFearSpell.cast(caster, target)
			endIf
		endIf
	endIf
	if (caster != Game.getPlayer())
		(self as ReferenceAlias).clear()
	endIf
	wasNaked = false
	thread_subject = None
endEvent

Function sendAlarm(Actor akTarget, Actor akCaster, bool naked = false)
	if (naked)
		wasNaked = true
	endIf
	if (akCaster == Game.getPlayer())
		if (akTarget.getCrimeFaction())
			previousCrimeGold = akTarget.getCrimeFaction().getCrimeGoldViolent()
		endIf
		if (akCaster.getFactionRank(SSLRSLoverFaction) < 2 && !wasNaked)
			;debug.messagebox("Alarming from "+akTarget.getLeveledActorBase().getName()+" to "+akCaster.getLeveledActorBase().getName())
			SSLRSCriminalPulseSpell.cast(akCaster, akTarget)
		endIf
	else
		if (!wasNaked && !caster.isGuard())
			;debug.messagebox("Alarming from "+akTarget.getLeveledActorBase().getName()+" to "+akCaster.getLeveledActorBase().getName())
			SSLRSCriminalPulseSpell.cast(akCaster, akTarget)
		endIf
	endIf
endFunction

Function stopAlarm()
	if (caster == Game.getPlayer())
		if (target.getCrimeFaction())
			target.getCrimeFaction().setCrimeGoldViolent(previousCrimeGold)
		endIf
	endIf
	caster.setAttackActorOnSight(false)
	caster.clearArrested()
	caster.stopCombatAlarm()
endFunction

Function storeArousal(Actor akActor)
	if (Game.getFormFromFile(0x3fc36, "SexLabAroused.esm"))
		previousArousal = getArousal(akActor)
	endIf
endFunction

float Function getArousal(Actor akActor)
	Quest slaFramework = Game.getFormFromFile(0x4290f, "SexLabAroused.esm") as Quest
	if (slaFramework)
		return (slaFramework as slaFrameworkScr).getActorArousal(akActor)
	else
		return 0.0
	endIf
endFunction

Function occupyCell(Actor akCaster, Actor akTarget)
	int belongings = akTarget.getNumItems()
	Form nthItem = None
	while (belongings > 0)
		belongings -= 1
		nthItem = akTarget.getNthForm(belongings)
		if (nthItem.getType() == 45 && akCaster.getItemCount(nthItem) == 0)
			showMessage(akTarget.getLeveledActorBase().getName()+" yielded to "+akCaster.getLeveledActorBase().getName())
			akCaster.additem(nthItem, 1)
		endIf
	endWhile
	Cell theRoom = akTarget.getParentCell()
	if (theRoom && theRoom.isInterior())
		ActorBase actorOwner = theRoom.getActorOwner()
		Faction factionOwner = theRoom.getFactionOwner()
		if (!factionOwner || !akCaster.isInFaction(factionOwner))
			if (akTarget.isInFaction(factionOwner))
				akCaster.addToFaction(factionOwner)
				showMessage(akCaster.getLeveledActorBase().getName()+" joined in ownership of "+theRoom.getName())
			elseIf (actorOwner && actorOwner == akTarget.getActorBase())
				factionOwner = SSLRSLoverFaction
				theRoom.setFactionOwner(factionOwner)
				akCaster.setFactionRank(factionOwner, 1)
				akTarget.setFactionRank(factionOwner, 0)
				showMessage(akCaster.getLeveledActorBase().getName()+" joined in ownership of "+theRoom.getName())
			endIf
		endIf
	endIf
	if (akCaster != Game.getPlayer() && akTarget == Game.getPlayer())
		akTarget.addPerk(SSLRSSexualInferiorityPerk)
		akCaster.setFactionRank(SSLRSLoverFaction, 2)
		showMessage("You are overwhelmed by "+akCaster.getLeveledActorBase().getName()+".", true)
	elseIf (akTarget != Game.getPlayer() && akCaster == Game.getPlayer())
		if (akTarget.getFactionRank(SSLRSLoverFaction) > 1)
			akCaster.removePerk(SSLRSSexualInferiorityPerk)
			if (akTarget.getFactionRank(SSLRSLoverFaction) == 2)
				akTarget.removeFromFaction(SSLRSLoverFaction)
			endIf
		endIf
	endIf
endFunction

Event onUpdate()
	if (target && target.is3DLoaded() && !target.isBleedingOut())
		debug.sendAnimationEvent(target, "BleedoutStop")
	endIf
endEvent

Event onHit(ObjectReference akAggressor, Form akSource, Projectile akProjectile, Bool abPowerAttack, Bool abSneakAttack, Bool abBashAttack, Bool abHitBlocked)
	Weapon source = akSource as Weapon
	if (thread_subject)
		if (source && source.getWeaponType() <= 6)
			interrupted = true
			thread_subject.endAnimation(true)
			getActorRef().stopCombatAlarm()
			(akAggressor as Actor).clearArrested()
		endIf
	endIf
endEvent

Function showMessage(String msg, bool staticMsg = false)
	if (caster == Game.getPlayer() || target == Game.getPlayer())
		if (staticMsg)
			debug.messageBox(msg)
		else
			debug.notification(msg)
		endIf
	endIf
endFunction
